
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <CL/opencl.h>

#define TIME_DIFF(t1, t2) \
  ((t2.tv_sec - t1.tv_sec) * 1000000 + (t2.tv_usec - t1.tv_usec))

#define MAX_DEVICES 5


#define error(...) do { fprintf(stderr, "Error: " __VA_ARGS__); exit(EXIT_FAILURE); } while(0)
#define check(err, ...)					\
  do {							\
    if(err != CL_SUCCESS) {				\
      fprintf(stderr, "(%d) Error: " __VA_ARGS__, err);	\
      exit(EXIT_FAILURE);				\
    }							\
  } while(0)

size_t file_size(const char *filename) {
	struct stat sb;
	if (stat(filename, &sb) < 0) {
		perror ("stat");
		abort ();
	}
	return sb.st_size;
}

char *
load(const char *filename) {
	FILE *f;
	char *b;
	size_t s;
	size_t r;
	s = file_size (filename);
	b = malloc (s+1);
	if (!b) {
		perror ("malloc");
		exit (1);
	}
	f = fopen (filename, "r");
	if (f == NULL) {
		perror ("fopen");
		exit (1);
	}
	r = fread (b, s, 1, f);
	if (r != 1) {
		perror ("fread");
		exit (1);
	}
	b[s] = '\0';
	return b;
}

/* Version CPU pour comparer le resultat */
void add_mat(float* C, const float* A, const float* B, unsigned int size_x, unsigned int size_y) {
  for(unsigned int i=0;i<size_x*size_y;i++)
    C[i]=A[i]+B[i];
}


int main(int argc, char** argv)
{

  cl_platform_id	pf[3];
  cl_uint nb_platforms = 0;
  cl_uint p = -1;

  cl_context context;                 // compute context
  cl_program program;                 // compute program
  cl_int err;                            // error code returned from api calls
    
  cl_device_id devices[MAX_DEVICES];
  cl_uint nb_devices = 0;

  cl_device_type device_type = CL_DEVICE_TYPE_GPU;

  cl_mem d_idata;                       // device memory used for first matrix
  cl_mem d_idata2;                      // device memory used for second matrix
  cl_mem d_odata;                       // device memory used for result matrix
  cl_int dev;

  const unsigned int size_x = 2048;
  const unsigned int size_y = 2048;
  const unsigned int mem_size= size_x*size_y*sizeof(float);

  float *h_idata = NULL;
  float *h_idata2 = NULL;
  float *h_odata = NULL;
  unsigned int dim = 16; // Default workgroup size

  struct timeval tv1,tv2;

  // Filter args
  //
  argv++;
  while (argc > 1) {
    if(!strcmp(*argv, "--all")) {
      if(device_type != CL_DEVICE_TYPE_GPU)
	error("--all and --cpu-only can not be specified at the same time\n");
      device_type = CL_DEVICE_TYPE_ALL;
    } else if(!strcmp(*argv, "--cpu-only")) {
      if(device_type != CL_DEVICE_TYPE_GPU)
	error("--gpu-only and --cpu-only can not be specified at the same time\n");
      device_type = CL_DEVICE_TYPE_CPU;
    } else
      break;
    argc--; argv++;
  }

  if(argc > 1)
    dim = atoi(*argv);

  // Allocation of input & output matrices
  //
  h_idata = malloc(mem_size);
  h_idata2 = malloc(mem_size);
  h_odata = malloc(mem_size);

  // Initialization of input matrices
  //
  srand(1234);
  for(unsigned int i = 0; i < size_x * size_y; i++) {
    h_idata[i]=rand();
    h_idata2[i]=rand();
  }

  // Get list of OpenCL platforms detected
  //
  err = clGetPlatformIDs(3, pf, &nb_platforms);
  check(err, "Failed to get platform IDs");

  fprintf(stderr,"%d OpenCL platforms detected\n", nb_platforms);

  // Print name & vendor for each platform
  //                                                                                                                    
  for (unsigned int _p=0; _p<nb_platforms; _p++) {
    char name[1024], vendor[1024];

    err = clGetPlatformInfo(pf[_p], CL_PLATFORM_NAME, 1024, name, NULL);
    check(err, "Failed to get Platform Info");

    err = clGetPlatformInfo(pf[_p], CL_PLATFORM_VENDOR, 1024, vendor, NULL);
    check(err, "Failed to get Platform Info");

    fprintf(stderr,"Platform %d: %s - %s\n", _p, name, vendor);

    // Get list of devices
    //
    err = clGetDeviceIDs(pf[_p], device_type, MAX_DEVICES, devices, &nb_devices);
    if (err == CL_SUCCESS && nb_devices) {
      p = _p;
      fprintf(stderr,"Choosing platform %d\n", p);
      break;
    }
  }
  if (p == (cl_uint)-1) {
    fprintf(stderr,"No platform found !\n");
    exit(1);
  }

  // Get list of devices
  //
  err = clGetDeviceIDs(pf[p], device_type, MAX_DEVICES, devices, &nb_devices);
  fprintf(stderr,"nb devices = %d\n", nb_devices);

  // Create compute context with "device_type" devices
  //
  context = clCreateContext (0, nb_devices, devices, NULL, NULL, &err);
  check(err, "Failed to create compute context");

  // Load program source
  const char	*opencl_prog;
  opencl_prog = load("addmat.cl");

  // Build program
  //
  program = clCreateProgramWithSource(context, 1, &opencl_prog, NULL, &err);
  check(err, "Failed to create program");

  err = clBuildProgram (program, 0, NULL, NULL, NULL, NULL);
  check(err, "Failed to build program");

  // Create the input and output buffers in device memory for our calculation
  //
  d_idata = clCreateBuffer(context, CL_MEM_READ_ONLY, mem_size, NULL, NULL);
  if (!d_idata)
    error("Failed to allocate device memory!\n");

  d_idata2 = clCreateBuffer(context, CL_MEM_READ_ONLY, mem_size, NULL, NULL);
  if (!d_idata2)
    error("Failed to allocate device memory!\n");

  d_odata = clCreateBuffer(context, CL_MEM_WRITE_ONLY, mem_size, NULL, NULL);
  if (!d_odata)
    error("Failed to allocate device memory!\n");

  // TODO: mettre à zéro la sortie, pour être sûr
  // TODO: trouver à afficher l'erreur de compil

  // Iterate over devices
  //
  for(dev = 0; dev < nb_devices; dev++) {
    cl_command_queue queue;
    cl_kernel kernel;

    char name[1024];
    err = clGetDeviceInfo(devices[dev], CL_DEVICE_NAME, 1024, name, NULL);
    check(err, "Cannot get type of device");

    fprintf(stderr,"Device %d : [%s]\n", dev, name);

    // Create a command queue
    //
    queue = clCreateCommandQueue(context, devices[dev], CL_QUEUE_PROFILING_ENABLE, &err);
    check(err,"Failed to create a command queue!\n");

    // Here, we can distinguish between CPU and GPU devices so as
    // to use different kernels, different work group size, etc.
    {
      size_t global[2];                      // global domain size for our calculation
      size_t local[2];                       // local domain size for our calculation

      // Create the compute kernel in the program we wish to run
      //
      kernel = clCreateKernel(program, "add_mat_gpu", &err);
      check(err, "Failed to create compute kernel!\n");

      // Write our data sets into the device memory
      //
      err = clEnqueueWriteBuffer(queue, d_idata, CL_TRUE, 0,
				 mem_size, h_idata, 0, NULL, NULL);
      check(err, "Failed to transfer input matrix!\n");

      err = clEnqueueWriteBuffer(queue, d_idata2, CL_TRUE, 0,
                                 mem_size, h_idata2, 0, NULL, NULL);
      check(err, "Failed to transfer input matrix!\n");

      // Set the arguments to our compute kernel
      //
      err = 0;
      err |= clSetKernelArg(kernel, 0, sizeof(cl_mem), &d_odata);
      err |= clSetKernelArg(kernel, 1, sizeof(cl_mem), &d_idata);
      err |= clSetKernelArg(kernel, 2, sizeof(cl_mem), &d_idata2);
      err |= clSetKernelArg(kernel, 3, sizeof(unsigned int), &size_x);
      err |= clSetKernelArg(kernel, 4, sizeof(unsigned int), &size_y);
      check(err, "Failed to set kernel arguments! %d\n", err);

      global[0] = size_x;
      global[1] = size_y;
      local[0] = dim; // Set workgroup size to dim
      local[1] = 1;

      int numIterations = 200;

      gettimeofday(&tv1, NULL);
      for(int i = 0; i<numIterations; i++) // Iterations are done inside the kernel
      {
	err = clEnqueueNDRangeKernel(queue, kernel, 2, NULL, global, local, 0, NULL, NULL);
	check(err, "Failed to execute kernel!\n");

	// Wait for the command commands to get serviced before reading back results
	//
	clFinish(queue);
      }
      gettimeofday(&tv2, NULL);
      float time1=((float)TIME_DIFF(tv1,tv2)) / 1000;

      // Read back the results from the device to verify the output
      //
      err = clEnqueueReadBuffer(queue, d_odata, CL_TRUE, 0,
				mem_size, h_odata, 0, NULL, NULL );  
      check(err, "Failed to read output matrix! %d\n", err);

      /* Version cpu pour comparaison */
      float* reference = (float*) malloc(mem_size);
      gettimeofday(&tv1,NULL);

      for(int i=0;i<numIterations;++i) {
	add_mat(reference, h_idata, h_idata2, size_x, size_y);
      }
      gettimeofday(&tv2,NULL);
      float time2=((float)TIME_DIFF(tv1,tv2)) / 1000;

      printf("%d\t%f\t%f ms (%fGo/s)\t%f ms (%fGo/s)\n", dim, time2/time1,
	     time1, numIterations * 3*mem_size / time1 / 1000000,
	     time2, numIterations * 3*mem_size / time2 / 1000000);

      // Validate our results                                                                              
      //
      unsigned int errors=0;
      for(unsigned int i=0;i<size_x*size_y;i++){
	if(reference[i]!=h_odata[i])
	  errors++;
      }
      if(errors)
	fprintf(stderr,"%d erreurs !\n", errors);
      else
	fprintf(stderr,"pas d'erreurs, cool !\n");

      clReleaseKernel(kernel);

      free(reference);
    }

      clReleaseCommandQueue(queue);
  }

  // Shutdown and cleanup
  //
  free(h_odata);
  free(h_idata);
  free(h_idata2);
  clReleaseMemObject(d_odata);
  clReleaseMemObject(d_idata);
  clReleaseMemObject(d_idata2);
  clReleaseProgram(program);
  clReleaseContext(context);

  return 0;
}

