
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <CL/opencl.h>

#define TIME_DIFF(t1, t2) \
  ((t2.tv_sec - t1.tv_sec) * 1000000 + (t2.tv_usec - t1.tv_usec))

#define MAX_DEVICES 5


#define error(...) do { fprintf(stderr, "Error: " __VA_ARGS__); exit(EXIT_FAILURE); } while(0)
#define check(err, ...)					\
  do {							\
    if(err != CL_SUCCESS) {				\
      fprintf(stderr, "(%d) Error: " __VA_ARGS__, err);	\
      exit(EXIT_FAILURE);				\
    }							\
  } while(0)

size_t file_size(const char *filename) {
	struct stat sb;
	if (stat(filename, &sb) < 0) {
		perror ("stat");
		abort ();
	}
	return sb.st_size;
}

char *
load(const char *filename) {
	FILE *f;
	char *b;
	size_t s;
	size_t r;
	s = file_size (filename);
	b = malloc (s+1);
	if (!b) {
		perror ("malloc");
		exit (1);
	}
	f = fopen (filename, "r");
	if (f == NULL) {
		perror ("fopen");
		exit (1);
	}
	r = fread (b, s, 1, f);
	if (r != 1) {
		perror ("fread");
		exit (1);
	}
	b[s] = '\0';
	return b;
}

/* Version CPU pour comparer le resultat */
void heat(float* next, const float* previous, unsigned int size_x) {
  for(unsigned int i=0;i<size_x;i++)
    next[i+1] = (previous[i] + previous[i+1] * 2 + previous[i+2]) / 4;
}


int main(int argc, char** argv)
{

  cl_platform_id	pf[3];
  cl_uint nb_platforms = 0;
  cl_uint p = -1;

  cl_context context;                 // compute context
  cl_program program;                 // compute program
  cl_int err;                            // error code returned from api calls
    
  cl_device_id devices[MAX_DEVICES];
  cl_uint nb_devices = 0;

  cl_device_type device_type = CL_DEVICE_TYPE_GPU;

  cl_mem d_data1;                       // device memory used for first matrix
  cl_mem d_data2;                       // device memory used for result matrix
  cl_int dev;

  const unsigned int size_x = 4*1024*1024;
  const unsigned int mem_size= (size_x+2)*sizeof(float);

  float *h_data1 = NULL;
  float *h_data2 = NULL;
  unsigned int dim = 16; // Default workgroup size

  struct timeval tv1,tv2;

  // Filter args
  //
  argv++;
  while (argc > 1) {
    if(!strcmp(*argv, "--all")) {
      if(device_type != CL_DEVICE_TYPE_GPU)
	error("--all and --cpu-only can not be specified at the same time\n");
      device_type = CL_DEVICE_TYPE_ALL;
    } else if(!strcmp(*argv, "--cpu-only")) {
      if(device_type != CL_DEVICE_TYPE_GPU)
	error("--gpu-only and --cpu-only can not be specified at the same time\n");
      device_type = CL_DEVICE_TYPE_CPU;
    } else
      break;
    argc--; argv++;
  }

  if(argc > 1)
    dim = atoi(*argv);

  // Allocation of input & output matrices
  //
  h_data1 = malloc(mem_size);
  h_data2 = malloc(mem_size);

  // Initialization of input matrices
  //
  srand(1234);
  for(unsigned int i = 0; i < size_x+2; i++) {
    h_data1[i]=rand();
  }

  // Get list of OpenCL platforms detected
  //
  err = clGetPlatformIDs(3, pf, &nb_platforms);
  check(err, "Failed to get platform IDs");

  printf("%d OpenCL platforms detected\n", nb_platforms);

  // Print name & vendor for each platform
  //
  for (unsigned int _p=0; _p<nb_platforms; _p++) {
    char name[1024], vendor[1024];

    err = clGetPlatformInfo(pf[_p], CL_PLATFORM_NAME, 1024, name, NULL);
    check(err, "Failed to get Platform Info");

    err = clGetPlatformInfo(pf[_p], CL_PLATFORM_VENDOR, 1024, vendor, NULL);
    check(err, "Failed to get Platform Info");

    printf("Platform %d: %s - %s\n", _p, name, vendor);

    // Get list of devices
    //
    err = clGetDeviceIDs(pf[_p], device_type, MAX_DEVICES, devices, &nb_devices);
    if (err == CL_SUCCESS && nb_devices) {
      p = _p;
      printf("Choosing platform %d\n", p);
      break;
    }
  }
  if (p == (cl_uint)-1) {
    printf("No platform found !\n");
    exit(1);
  }

  // Get list of devices
  //
  err = clGetDeviceIDs(pf[p], device_type, MAX_DEVICES, devices, &nb_devices);
  printf("nb devices = %d\n", nb_devices);

  // Create compute context with "device_type" devices
  //
  context = clCreateContext (0, nb_devices, devices, NULL, NULL, &err);
  check(err, "Failed to create compute context");

  // Load program source
  const char	*opencl_prog;
  opencl_prog = load("heat.cl");

  // Build program
  //
  program = clCreateProgramWithSource(context, 1, &opencl_prog, NULL, &err);
  check(err, "Failed to create program");

  err = clBuildProgram (program, 0, NULL, NULL, NULL, NULL);
  check(err, "Failed to build program");

  // Create the input and output buffers in device memory for our calculation
  //
  d_data1 = clCreateBuffer(context, CL_MEM_READ_WRITE, mem_size, NULL, NULL);
  if (!d_data1)
    error("Failed to allocate device memory!\n");

  d_data2 = clCreateBuffer(context, CL_MEM_READ_WRITE, mem_size, NULL, NULL);
  if (!d_data2)
    error("Failed to allocate device memory!\n");

  // Iterate over devices
  //
  for(dev = 0; dev < nb_devices; dev++) {
    cl_command_queue queue;
    cl_kernel kernel;

    char name[1024];
    err = clGetDeviceInfo(devices[dev], CL_DEVICE_NAME, 1024, name, NULL);
    check(err, "Cannot get type of device");

    printf("Device %d : [%s]\n", dev, name);

    // Create a command queue
    //
    queue = clCreateCommandQueue(context, devices[dev], CL_QUEUE_PROFILING_ENABLE, &err);
    check(err,"Failed to create a command queue!\n");

    // Here, we can distinguish between CPU and GPU devices so as
    // to use different kernels, different work group size, etc.
    {
      size_t global[1];                      // global domain size for our calculation
      size_t local[1];                       // local domain size for our calculation

      // Create the compute kernel in the program we wish to run
      //
      kernel = clCreateKernel(program, "heat", &err);
      check(err, "Failed to create compute kernel!\n");

      // Write our data sets into the device memory
      //
      err = clEnqueueWriteBuffer(queue, d_data1, CL_TRUE, 0,
				 mem_size, h_data1, 0, NULL, NULL);
      check(err, "Failed to transfer input matrix!\n");

      global[0] = size_x;
      local[0] = dim; // Set workgroup size to dim

      int numIterations = 200;

      gettimeofday(&tv1, NULL);
      for(int i = 0; i<numIterations; i++) // Iterations are done inside the kernel
      {
	// Set the arguments to our compute kernel
	//
	err = 0;
	err |= clSetKernelArg(kernel, 0, sizeof(cl_mem), &d_data2);
	err |= clSetKernelArg(kernel, 1, sizeof(cl_mem), &d_data1);
	check(err, "Failed to set kernel arguments! %d\n", err);

	err = clEnqueueNDRangeKernel(queue, kernel, 1, NULL, global, local, 0, NULL, NULL);
	check(err, "Failed to execute kernel!\n");

	// Now the other way...
	//
	err = 0;
	err |= clSetKernelArg(kernel, 0, sizeof(cl_mem), &d_data1);
	err |= clSetKernelArg(kernel, 1, sizeof(cl_mem), &d_data2);
	check(err, "Failed to set kernel arguments! %d\n", err);

	err = clEnqueueNDRangeKernel(queue, kernel, 1, NULL, global, local, 0, NULL, NULL);
	check(err, "Failed to execute kernel!\n");

      }

      // Wait for the command commands to get serviced before reading back results
      //
      clFinish(queue);

      gettimeofday(&tv2, NULL);
      float time1=((float)TIME_DIFF(tv1,tv2)) / 1000;

      // Read back the results from the device to verify the output
      //
      err = clEnqueueReadBuffer(queue, d_data2, CL_TRUE, 0,
				mem_size, h_data2, 0, NULL, NULL );  
      check(err, "Failed to read output matrix! %d\n", err);

      /* Version cpu pour comparaison */
      float* reference = (float*) malloc(mem_size);
      gettimeofday(&tv1,NULL);

      for(int i=0;i<numIterations;++i) {
	heat(reference, h_data1, size_x);
	heat(h_data1, reference, size_x);
      }
     gettimeofday(&tv2,NULL);
      float time2=((float)TIME_DIFF(tv1,tv2)) / 1000;

      printf("%d\t%f\t%f ms (%fGo/s)\t%f ms (%fGo/s)\n", dim, time2/time1,
	     time1, 2 * numIterations * 4 * mem_size / time1 / 1000000,
	     time2, 2 * numIterations * 4 * mem_size / time2 / 1000000);

      /* Comparaison des resultats */
      unsigned int errors=0;
      for(unsigned int i=0;i<size_x;i++){
	if((reference[i]-h_data2[i])/reference[i] > 1e-6)
	  errors++;
      }
      if(errors)
	fprintf(stderr,"%d erreurs !\n", errors);
      else
	fprintf(stderr,"pas d'erreurs, cool !\n");

      clReleaseKernel(kernel);

      free(reference);
    }

      clReleaseCommandQueue(queue);
  }

  // Shutdown and cleanup
  //
  free(h_data1);
  free(h_data2);
  clReleaseMemObject(d_data1);
  clReleaseMemObject(d_data2);
  clReleaseProgram(program);
  clReleaseContext(context);

  return 0;
}

