
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <CL/opencl.h>

#define MAX_DEVICES 5

// Use a static data size for simplicity
//
#define DATA_SIZE  (1024*1024)
#define CHUNK_SIZE 16

#define error(...) do { fprintf(stderr, "Error: " __VA_ARGS__); exit(EXIT_FAILURE); } while(0)
#define check(err, ...)					\
  do {							\
    if(err != CL_SUCCESS) {				\
      fprintf(stderr, "(%d) Error: " __VA_ARGS__, err);	\
      exit(EXIT_FAILURE);				\
    }							\
  } while(0)

size_t file_size(const char *filename) {
	struct stat sb;
	if (stat(filename, &sb) < 0) {
		perror ("stat");
		abort ();
	}
	return sb.st_size;
}

char *
load(const char *filename) {
	FILE *f;
	char *b;
	size_t s;
	size_t r;
	s = file_size (filename);
	b = malloc (s+1);
	if (!b) {
		perror ("malloc");
		exit (1);
	}
	f = fopen (filename, "r");
	if (f == NULL) {
		perror ("fopen");
		exit (1);
	}
	r = fread (b, s, 1, f);
	if (r != 1) {
		perror ("fread");
		exit (1);
	}
	b[s] = '\0';
	return b;
}

int main(int argc, char** argv)
{

  cl_platform_id	pf[3];
  cl_uint nb_platforms = 0;
  cl_uint p = -1;

  cl_context context;                 // compute context
  cl_program program;                 // compute program
  cl_int err;                            // error code returned from api calls
    
  cl_device_id devices[MAX_DEVICES];
  cl_uint nb_devices = 0;

  cl_device_type device_type = CL_DEVICE_TYPE_GPU;

  cl_mem input;                       // device memory used for the input array
  cl_int dev;

  float *data;                        // original data set given to device
  unsigned int correct;               // number of correct results returned

  // Filter args
  //
  argv++;
  while (argc > 1) {
    if(!strcmp(*argv, "--all")) {
      if(device_type != CL_DEVICE_TYPE_GPU)
	error("--all and --cpu-only can not be specified at the same time\n");
      device_type = CL_DEVICE_TYPE_ALL;
    } else if(!strcmp(*argv, "--cpu-only")) {
      if(device_type != CL_DEVICE_TYPE_GPU)
	error("--gpu-only and --cpu-only can not be specified at the same time\n");
      device_type = CL_DEVICE_TYPE_CPU;
    } else
      break;
    argc--; argv++;
  }

  data = malloc(DATA_SIZE * sizeof(float));

  // Get list of OpenCL platforms detected
  //
  err = clGetPlatformIDs(3, pf, &nb_platforms);
  check(err, "Failed to get platform IDs");

  printf("%d OpenCL platforms detected\n", nb_platforms);

  // Print name & vendor for each platform
  //
  for (unsigned int _p=0; _p<nb_platforms; _p++) {
    char name[1024], vendor[1024];

    err = clGetPlatformInfo(pf[_p], CL_PLATFORM_NAME, 1024, name, NULL);
    check(err, "Failed to get Platform Info");

    err = clGetPlatformInfo(pf[_p], CL_PLATFORM_VENDOR, 1024, vendor, NULL);
    check(err, "Failed to get Platform Info");

    printf("Platform %d: %s - %s\n", _p, name, vendor);

    // Get list of devices
    //
    err = clGetDeviceIDs(pf[_p], device_type, MAX_DEVICES, devices, &nb_devices);
    if (err == CL_SUCCESS && nb_devices) {
      p = _p;
      printf("Choosing platform %d\n", p);
      break;
    }
  }
  if (p == (cl_uint)-1) {
    printf("No platform found !\n");
    exit(1);
  }

  // Get list of devices
  //
  err = clGetDeviceIDs(pf[p], device_type, MAX_DEVICES, devices, &nb_devices);
  printf("nb devices = %d\n", nb_devices);

  // Create compute context with "device_type" devices
  //
  context = clCreateContext (0, nb_devices, devices, NULL, NULL, &err);
  check(err, "Failed to create compute context");

  // Load program source
  const char	*opencl_prog;
  opencl_prog = load("vector.cl");

  // Build program
  //
  program = clCreateProgramWithSource(context, 1, &opencl_prog, NULL, &err);
  check(err, "Failed to create program");

  err = clBuildProgram (program, 0, NULL, NULL, NULL, NULL);
  check(err, "Failed to build program");

  // Create the input and output arrays in device memory for our calculation
  //
  input = clCreateBuffer(context,  CL_MEM_READ_WRITE,  sizeof(float) * DATA_SIZE, NULL, NULL);
  if (!input)
    error("Failed to allocate device memory!\n");
    
  // Iterate over devices
  //
  for(dev = 0; dev < nb_devices; dev++) {
    cl_command_queue queue;
    cl_kernel kernel;

    char name[1024];
    err = clGetDeviceInfo(devices[dev], CL_DEVICE_NAME, 1024, name, NULL);
    check(err, "Cannot get type of device");

    printf("Device %d : [%s]\n", dev, name);

    // Fill input data set
    //
    int i = 0;
    for(i = 0; i < DATA_SIZE; i++)
      data[i] = (float)i;

    // Create a command queue
    //
    queue = clCreateCommandQueue(context, devices[dev], CL_QUEUE_PROFILING_ENABLE, &err);
    check(err,"Failed to create a command queue!\n");

    // Here, we can distinguish between CPU and GPU devices so as
    // to use different kernels, different work group size, etc.
    {
      cl_event prof_event;
      size_t global;                      // global domain size for our calculation
      size_t local;                       // local domain size for our calculation

      // Create the compute kernel in the program we wish to run
      //
      kernel = clCreateKernel(program, "square", &err);
      check(err, "Failed to create compute kernel!\n");

      // Write our data set into the input array in device memory 
      //
      err = clEnqueueWriteBuffer(queue, input, CL_TRUE, 0,
				 sizeof(float) * DATA_SIZE, data, 0, NULL, NULL);
      check(err, "Failed to write to input array!\n");

      // Set the arguments to our compute kernel
      //
      err = 0;
      err  = clSetKernelArg(kernel, 0, sizeof(cl_mem), &input);
      check(err, "Failed to set kernel arguments! %d\n", err);

	global = DATA_SIZE;
	local = 16; // Set workgroup size to 16

	err = clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global, &local, 0, NULL, &prof_event);
	check(err, "Failed to execute kernel!\n");

	// Wait for the command commands to get serviced before reading back results
	//
	clFinish(queue);

	// Read back the results from the device to verify the output
	//
	err = clEnqueueReadBuffer(queue, input, CL_TRUE, 0,
				  sizeof(float) * DATA_SIZE, data, 0, NULL, NULL );  
	check(err, "Failed to read output array! %d\n", err);
    
	// Check performance
	//
	{
	  cl_ulong start, end;
	  float timeInMicroseconds;

	  clGetEventProfilingInfo(prof_event, CL_PROFILING_COMMAND_START,
				  sizeof(cl_ulong), &start, NULL);

	  clGetEventProfilingInfo(prof_event, CL_PROFILING_COMMAND_END,
				  sizeof(cl_ulong), &end, NULL);

	  timeInMicroseconds = (end - start) * 1.0e-3f;

	  printf("Computation performed in %f µs over device #%d\n",
		 timeInMicroseconds,
		 dev);

	  clReleaseEvent(prof_event);
	}

	// Validate our results
	//
	correct = 0;
	for(i = 0; i < DATA_SIZE; i++) {
	  float expected = (float)((i / CHUNK_SIZE)*CHUNK_SIZE + (CHUNK_SIZE-1 - i % CHUNK_SIZE));

	  if(data[i] == expected)
	    correct++;
	  if(i < 2*CHUNK_SIZE)
	    printf("results[%d] = %f\n", i, data[i]);

	}
	printf("Computed '%d/%d' correct values!\n", correct, DATA_SIZE);
    
	clReleaseKernel(kernel);
      }

      clReleaseCommandQueue(queue);
    }

    // Shutdown and cleanup
    //
    clReleaseMemObject(input);
    clReleaseProgram(program);
    clReleaseContext(context);

    return 0;
}

